#ifndef __ARRAY_H
#define __ARRAY_H

#include <SophiaFramework.hpp>

// Cӂ̌^AIzNX

#define ARRAY_ALLOCATE_STEP 200     // mۂ郁̍ŏP(g܂)

template <typename T>
class Array {
private:
    Bool Expand(SInt32 size) {
        // zg
        T *p = NewArray(size);
        if (p == null) {
            return false;
        }
        if (_number > 0) {
            for (SInt32 i = 0; i < _number; i++) {
                p[i] = _buffer[i];
            }
        }
        if (_buffer != null) {
            DeleteArray(_buffer, _size);
        }
        _buffer = p;
        _size = size;
        return true;
    }

    Bool Shrink(SInt32 size) {
        // zk
        T *p = NewArray(size);
        if (p == null) {
            return false;
        }
        if (_number > 0) {
            for (SInt32 i = 0; i < size; i++) {
                p[i] = _buffer[i];
            }

        }
        if (_buffer != null) {
            DeleteArray(_buffer, _size);
        }
        _buffer = p;
        if (_number > size) {
            _number = size;
        }
        _size = size;
        return true;
    }

    // Ǝ̃Ǘ@\(new)
    T *NewArray(SInt32 size) {
        T *p = (T *)MemoryAllocate(sizeof(T) * size);
        if (p != null) {
            for (SInt32 i = 0; i < size; i++) {
                ::new(p + i) T();
            }
        }
        return p;
    }

    // Ǝ̃Ǘ@\(delete)
    Void DeleteArray(T *p, SInt32 size) {
    if (p != null) {
            for (SInt32 i = 0; i < size; i++) {
                T *q = p + i;
                q->~T();
            }
            MemoryFree(p);
            p = null;
        }
    }

protected:
    T *_buffer;
    SInt32 _number;
    SInt32 _size;

public:
    // RXgN^
    Array(Void) {
        _buffer = null;
        _number = 0;
        _size = 0;
    }

    // Rs[RXgN^
    Array(const Array<T> &obj) {
        _buffer = null;
        _number = 0;
        _size = 0;
        Set(obj);
    }

    // fBXgN^
    ~Array(Void) {
        Clear();
    }

    // eSĔj
    Void Clear(Void) {
        if (_buffer != null) {
            DeleteArray(_buffer, _size);
        }
        _buffer = null;
        _number = 0;
        _size = 0;
    }

    // IuWFNgRs[
    SFCError Set(const Array<T> &obj) {
        if (obj._number == 0) {
            Clear();
            return SFERR_NO_ERROR;
        }
        if (SetSize(obj._number) != SFERR_NO_ERROR) {
            return SFERR_NO_MEMORY;
        }
        for (SInt32 i = 0; i < obj._number; i++) {
            _buffer[i] = obj._buffer[i];
        }
        _number = obj._number;
        return SFERR_NO_ERROR;
    }

    // vfݒ
    SFCError Set(SInt32 index, const T &obj) {
        ASSERT(index >= 0 && index < _number);
        _buffer[index] = obj;
        return SFERR_NO_ERROR;
    }

    // vf擾
    const T &Get(SInt32 index) const {
        ASSERT(index >= 0 && index < _number);
        return _buffer[index];
    }

    // z̃TCYݒ
    SFCError SetSize(SInt32 size) {
        ASSERT (size >= 0);
        if (size == 0) {
            Clear();
            return SFERR_NO_ERROR;
        }
        if (size > _size) {
            Bool ret(false);
            if ((size - _number) < ARRAY_ALLOCATE_STEP) {
                ret = Expand(_number + ARRAY_ALLOCATE_STEP);
            } else {
                ret = Expand(size);
            }
            if (ret) {
                _number = size;
            }
            return ret ? SFERR_NO_ERROR : SFERR_NO_MEMORY;
        } else {
            Bool ret(true);
            if ((_size - size) >= ARRAY_ALLOCATE_STEP) {
                ret = Shrink(size);
            }
            _number = size;
            return ret ? SFERR_NO_ERROR : SFERR_NO_MEMORY;
        }
    }

    // vfԂ
    SInt32 GetSize(Void) const {
        return _number;
    }

    // z񂪋󂩂ǂ
    Bool IsEmpty(Void) const {
        if (_number == 0) {
            return true;
        } else {
            return false;
        }
    }

    // vf𖖔ɒǉ
    SFCError Append(const T &obj) {
        SInt32Const oldNumber(_number);
        if(SetSize(_number + 1) != SFERR_NO_ERROR) {    // ̊mۂɎs
            return SFERR_NO_MEMORY;
        }
        _buffer[oldNumber] = obj;
        return SFERR_NO_ERROR;
    }

    // vfQ𖖔ɒǉ
    SFCError Append(const Array<T> &obj) {
        SInt32Const oldNumber(_number);
        if (SetSize(_number + obj._size) != SFERR_NO_ERROR) {
            return SFERR_NO_MEMORY;
        }
        for (SInt32 i = 0; i < obj._number; i++) {
            _buffer[oldNumber + i] = obj._buffer[i];
        }
        return SFERR_NO_ERROR;
    }

    // vf𖖔ɒǉ
    SFCError InsertLast(const T &obj) {
        return Append(obj);
    }

    // vfQ𖖔ɒǉ
    SFCError InsertLast(const Array<T> &obj) {
        return Append(obj);
    }

    // vf}
    SFCError Insert(SInt32 index, const T &obj) {
        ASSERT(index >= 0 && index < _number);
        SInt32 i(_number);
        if (SetSize(_number + 1) != SFERR_NO_ERROR) {
            return SFERR_NO_MEMORY;
        }
        while (i > index) {
            _buffer[i] = _buffer[--i];
        }
        _buffer[i] = obj;
        return SFERR_NO_ERROR;
    }

    // vfQ}
    SFCError Insert(SInt32 index, const Array<T> &obj) {
        ASSERT(index >=0 && index < _number);
        if (obj.IsEmpty()) {
            return SFERR_NO_ERROR;
        }
        SInt32Const oldNumber(_number);
        if (SetSize(_number + obj._number) != SFERR_NO_ERROR) {
            return SFERR_NO_MEMORY;
        }
        SInt32Const n(oldNumber - index);
        SInt32 i;
        for (i = 0; i < n; i++) {
            SInt32 j = oldNumber - i - 1;
            _buffer[j + obj._number] = _buffer[j];
        }
        for (i = 0; i < obj._number; i++) {
            _buffer[i + index] = obj._buffer[i];
        }
        return SFERR_NO_ERROR;
    }

    // vf擪ɒǉ
    SFCError InsertFirst(const T &obj) {
        return Insert(0, obj);
    }

    // vfQ擪ɒǉ
    SFCError InsertFirst(const Array<T> &obj) {
        return Insert(0, obj);
    }

    // 擪̗vfԂ
    const T &GetFirst(Void) const {
        return Get(0);
    }

    // ̗vfԂ
    const T &GetLast(Void) const {
        return Get(_number - 1);
    }

    // ̗vf폜
    Void RemoveLast(Void) {
        SetSize(_number - 1);
    }

    Array<T> &operator=(const Array<T> &obj) {
        Set(obj);
        return *this;
    }

    const T &operator[](SInt32 index) const {
        return Get(index);
    }

    T &operator[](SInt32 index) {
        ASSERT(index >= 0 && index < _number);
        return _buffer[index];
    }
};

#endif
