#include "Bitmap.hpp"

Bitmap::Bitmap(Void) : _pbitmap(null), _pdib(null)
{
}

Bitmap::Bitmap(ACharConstPtr file) : _pbitmap(null), _pdib(null)
{
    LoadBitmap(file);
}

Bitmap::Bitmap(WCharConstPtr file) : _pbitmap(null), _pdib(null)
{
    LoadBitmap(file);
}

Bitmap::Bitmap(SFXAnsiStringConstRef file) : _pbitmap(null), _pdib(null)
{
    LoadBitmap(file);
}

Bitmap::Bitmap(SFXWideStringConstRef file) : _pbitmap(null), _pdib(null)
{
    LoadBitmap(file);
}

Bitmap::Bitmap(ACharConstPtr resFile, UInt16 id) : _pbitmap(null), _pdib(null)
{
    LoadResBitmap(resFile, id);
}

Bitmap::Bitmap(WCharConstPtr resFile, UInt16 id) : _pbitmap(null), _pdib(null)
{
    LoadResBitmap(resFile, id);
}

Bitmap::Bitmap(SFXAnsiStringConstRef resFile, UInt16 id) : _pbitmap(null), _pdib(null)
{
    LoadResBitmap(resFile, id);
}

Bitmap::Bitmap(SFXWideStringConstRef resFile, UInt16 id) : _pbitmap(null), _pdib(null)
{
    LoadResBitmap(resFile, id);
}

Bitmap::Bitmap(SFBBitmapSmpConstRef pbitmap) : _pbitmap(null), _pdib(null)
{
    Create(pbitmap);
}

Bitmap::Bitmap(BitmapConstRef bitmap) : _pbitmap(null), _pdib(null)
{
    Copy(bitmap);
}

Bitmap::~Bitmap(Void)
{
    _pbitmap.Release();
    _pdib.Release();
}

// DIB `̃rbg}bv擾
SFCError Bitmap::GetDib(Void)
{
    SFCError error(SFERR_NO_ERROR);
    _pdib.Release();
    if(_pbitmap != null) {
        SFBDIBPtr pdib;
        error = _pbitmap->QueryInterface(AEECLSID_BITMAP, reinterpret_cast<VoidHandle>(&pdib));
        if (error == SFERR_NO_ERROR) {
            _pdib.Attach(pdib);
        } else {
            _pbitmap.Release();
        }
    } else {
        error = SFERR_FAILED;
    }
    return error;
}

SFCError Bitmap::LoadBitmap(ACharConstPtr file)
{
    _pbitmap = SFBShell::GetInstance()->LoadBitmap(file);
    return GetDib();
}

SFCError Bitmap::LoadBitmap(WCharConstPtr file)
{
    SFXAnsiString ansiFile(file);
    return LoadBitmap(ansiFile);
}

SFCError Bitmap::LoadBitmap(SFXAnsiStringConstRef file)
{
    _pbitmap = SFBShell::GetInstance()->LoadBitmap(file);
    return GetDib();
}

SFCError Bitmap::LoadBitmap(SFXWideStringConstRef file)
{
    SFXAnsiString ansiFile(file);
    return LoadBitmap(ansiFile);
}

SFCError Bitmap::LoadResBitmap(ACharConstPtr resFile, UInt16 id)
{
    _pbitmap = SFBShell::GetInstance()->LoadResBitmap(resFile, id);
    return GetDib();
}

SFCError Bitmap::LoadResBitmap(WCharConstPtr resFile, UInt16 id)
{
    SFXAnsiString ansiResFile(resFile);
    return LoadResBitmap(ansiResFile, id);
}

SFCError Bitmap::LoadResBitmap(SFXAnsiStringConstRef resFile, UInt16 id)
{
    _pbitmap = SFBShell::GetInstance()->LoadResBitmap(resFile, id);
    return GetDib();
}

SFCError Bitmap::LoadResBitmap(SFXWideStringConstRef resFile, UInt16 id)
{
    SFXAnsiString ansiResFile(resFile);
    return LoadResBitmap(ansiResFile, id);
}

// SFBBitmapSmp ̓eRs[
SFCError Bitmap::Create(SFBBitmapSmpConstRef pbitmap)
{
    SFCError error(SFERR_NO_ERROR);
    // rbg}bv̑傫擾
    AEEBitmapInfo bmpinfo;
    pbitmap->GetInfo(&bmpinfo);
    // rbg}bvRs[
    SFBBitmapSmp tmpBmp;
    SFBBitmapSmp disp(SFXGraphics::GetInstance()->GetDestination());
    error = disp->CreateCompatibleBitmap(&tmpBmp, static_cast<UInt16>(bmpinfo.cx), static_cast<UInt16>(bmpinfo.cy));
    if (error == SFERR_NO_ERROR) {
        error = tmpBmp->BltIn(SFXRectangle(0, 0, static_cast<SInt16>(bmpinfo.cx), static_cast<SInt16>(bmpinfo.cy)), pbitmap, SFXGrid(0, 0));
        if (error == SFERR_NO_ERROR) {
            _pbitmap = tmpBmp;
            error = GetDib();
        }
    }
    return error;
}

// Bitmap IuWFNgRs[
SFCError Bitmap::Copy(BitmapConstRef bitmap)
{
    return Create(bitmap._pbitmap);
}

BitmapConstRef Bitmap::operator =(SFBBitmapSmpConstRef pbitmap)
{
    Create(pbitmap);
    return *this;
}

BitmapConstRef Bitmap::operator =(BitmapConstRef bitmap)
{
    Copy(bitmap);
    return *this;
}

SFXRGBColor Bitmap::GetPixel(SFXPixelConstRef pixel) const
{
    if (_pdib == null) {
        return SFXRGBColor(0, 0, 0, 0);
    }

    UInt32 bitsIndex(_pdib->GetDepth());
    bitsIndex *= pixel.GetX();
    BytePtr p(_pdib->GetBuffer() + pixel.GetY() * _pdib->GetPitch() + (bitsIndex >> 3));
                // bitsIndex >> 3  bitsIndex / 8 Ɠ
                // ŁAw肳ꂽ_̒li[Ă擪̃ AhX擾ł
    SInt08 localBitsIndex(8 - static_cast<SInt08>(bitsIndex & 7));
                // bitsIndex & 7  bitsIndex % 8 Ɠ
                //  p ̉牽rbgڂɖړĨf[^̂擾ł
    SInt16 len(_pdib->GetDepth());
    NativeColor color(0);
    BytePtr pcolor(reinterpret_cast<BytePtr>(&color));
    SInt08 leftBits(8);     // pcolor ɂƉrbg߂邩
    UInt08 getbits;
    UInt16 mask;

    while (len > 0) {
        // len  localBitsIndex  leftBits ̂ŏ̂̂ getbits ɑ
        getbits = (localBitsIndex < leftBits) ? localBitsIndex : leftBits;
        if (getbits > len) {
            getbits = static_cast<UInt08>(len);
        }

        mask = static_cast<UInt16>(1 << localBitsIndex) - 1;
        *pcolor = (*pcolor << getbits) | ((static_cast<UInt08>(mask) & *p) >> ((getbits < localBitsIndex) ? (localBitsIndex - getbits) : 0));
        len -= getbits;
        leftBits -= getbits;
        localBitsIndex -= getbits;
        if (localBitsIndex <= 0) {
            ++p;
            localBitsIndex = 8;
        }
        if (leftBits <= 0) {
            ++pcolor;
            leftBits = 8;
        }
    }
    return _pdib->NativeToColor(color);
}

SFXSize Bitmap::GetSize(Void) const
{
    if (_pdib == null) {
        return SFXSize(0, 0);
    } else {
        return _pdib->GetSize();
    }
}

UInt16 Bitmap::GetWidth(Void) const
{
    if (_pdib == null) {
        return 0;
    } else {
        return _pdib->GetWidth();
    }
}

UInt16 Bitmap::GetHeight(Void) const
{
    if (_pdib == null) {
        return 0;
    } else {
        return _pdib->GetHeight();
    }
}

UInt08 Bitmap::GetDepth(Void) const
{
    if (_pdib == null) {
        return 0;
    } else {
        return _pdib->GetDepth();
    }
}
