#include "Camera.hpp"

Camera::Camera(Void) : _camera(null), _bitmap(null), _previewable(false), _recordable(false), _spp(null)
{
    _mediaData.clsData = MMD_FILE_NAME;
    _mediaData.pData = "camera.jpg";
    _mediaData.dwSize = 0;
    _mimeType = "image/jpeg";
    return;
}

Camera::~Camera(Void)
{
    // fBXgN^
    Terminate();
    return;
}

// rect         : vr[\̈
// spp          : R[obN֐ւ̃|C^
// reference    : R[obN֐֓nf[^ւ̃|C^
SFCError Camera::Initialize(SFXRectangleConstRef rect, NotifySPP spp, VoidPtr reference)
{

    SFCError error(SFERR_NO_ERROR);
    _bitmap = SFXGraphics::CreateBitmap(rect.GetSize());    // J摜ێprbg}bv쐬
    if (_bitmap != null) {
        error = Resume();   // J̊{Iȏ
        if (error == SFERR_NO_ERROR) {
            _rect = rect;
            _spp = spp;
            _reference = reference;
        }
    } else {
        error = SFERR_NO_MEMORY;        // ŝ߂Ƀrbg}bv̍쐬oȂ
    }
    if (error != SFERR_NO_ERROR) {
        Terminate();
    }
    return error;
}

// vr[Jn
SFCError Camera::Preview(Void)
{
    SFCError error(SFERR_NO_ERROR);
    if (_camera != null) {
        _status = STATUS_PREVIEW;
        _previewable = false;
        _recordable = false;
        error = _camera->Preview();     // Jvr[[hŋN
    } else {
        error = SFERR_FAILED;
    }
    return error;
}

// XibvVbg[hŎBe
SFCError Camera::Record(Void)
{
    SFCError error(SFERR_NO_ERROR);
    if (_camera != null) {
        _previewable = false;
        _recordable = false;

        _status = STATUS_RECORD;
        error = _camera->Stop();    // vr[I
                                    // vr[ۂɏI_ŃCxgnhĂ΂̂ŁA
                                    // XibvVbg[hł̎Be͂̒ōsĂ
    } else {
        error = SFERR_FAILED;
    }
    return error;
}

// J̏I
Void Camera::Suspend(Void)
{
    if (_camera != null) {
        _previewable = false;
        _recordable = false;
        _camera->Stop();
        _camera.Release();
    }
    return;
}

// J̊{Iȏs
SFCError Camera::Resume(Void)
{
    SFCError error(SFERR_NO_ERROR);
    if ((_camera = SFBCamera::NewInstance()) != null) {     // J̃CX^X擾
        _previewable = true;
        _recordable = false;
        error = _camera->RegisterNotify(OnCameraSHP, this); // R[obNGgo^
        if (error == SFERR_NO_ERROR) {
            AEESize size;
            size.cx = PICTURE_WIDTH;
            size.cy = PICTURE_HEIGHT;
            error = _camera->SetSize(&size);
            if (error == SFERR_NO_ERROR) {
                error = _camera->SetDisplaySize(&size);         // JŎ擾摜̃TCYݒ
            }
        }
    } else {
        error = SFERR_NO_MEMORY;    // ŝ߂ɃJ̃CX^X擾łȂ
    }
    if (error != SFERR_NO_ERROR) {
        Suspend();
    }
    return error;
}

// vr[IāAg\[X̌n
Void Camera::Terminate(Void)
{
    Suspend();
    if (_bitmap != null) {
        _bitmap.Release();  // ŏɍ쐬rbg}bv
    }
    return;
}

// Xibv VbgBȅ
Void Camera::AfterSnapshot(Void)
{
    SFBBitmapSmp bitmap;
    if (_camera->GetFrame(&bitmap) == SFERR_NO_ERROR) { // J̉摜rbg}bvƂ
        _bitmap->BltIn(SFXRectangle(0, 0, _rect.GetSize()), bitmap, _rect.GetOrigin()); // J̉摜w肵TCYo
    }
    _previewable = true;
    _spp(_status, SFERR_NO_ERROR, _reference);
}

// R[obNnh(_~[)
Void OnCameraSHP(VoidPtr reference, AEECameraNotify *notify)
{
    CameraPtr(reference)->OnCamera(notify); // R[obNnh̎̂Ăяo
    return;
}

// R[obNnh()
Void Camera::OnCamera(AEECameraNotify *notify)
{
    switch (notify->nStatus) {
        case CAM_STATUS_START:  // J̋N
            if (_status == STATUS_PREVIEW) {
                _recordable = true;
                _spp(_status, SFERR_NO_ERROR, _reference);
            }
            break;
        case CAM_STATUS_FRAME:  // Vȃt[擾
            if (_status == STATUS_PREVIEW) {
                SFBBitmapSmp bitmap;
                if (_camera->GetFrame(&bitmap) == SFERR_NO_ERROR) {     // J̉摜rbg}bvƂĎ擾
                    _bitmap->BltIn(SFXRectangle(0, 0, _rect.GetSize()), bitmap, _rect.GetOrigin()); // J̉摜ʕ\̈𔲂o
                }
                _spp(STATUS_FRAME, SFERR_NO_ERROR, _reference);

                SFXGraphicsPtr pg = SFXGraphics::GetInstance();
                pg->BitBlt(_rect, _bitmap, SFXGrid(0, 0));      // 擾摜ʂɕ\
                pg->Update();       // ʂXV
            }
            break;
        case CAM_STATUS_DONE:   // 
            if (_status == STATUS_RECORD) {
                _status = STATUS_ENCODE;
                _camera->DeferEncode(true); // xGR[h[hɐݒ
                _camera->SetMediaData(&_mediaData, _mimeType);
                _camera->RecordSnapshot();  // XibvVbg[hŎBe
            } else if (_status == STATUS_ENCODE) {
                AfterSnapshot();
            }
            break;
        case CAM_STATUS_FAIL:
        case CAM_STATUS_ABORT:
            _spp(_status, EFAILED, _reference);
            break;
    }
    return;
}

// ݂̖邳擾
SInt16 Camera::GetBrightness(Void)
{
    SInt16 result(0);
    if (_camera != null) {
        SInt32 value;
        AEEParmInfo info;
        if (_camera->GetParm(CAM_PARM_BRIGHTNESS, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {    // 邳擾
            result = (SInt16)value;
        }
    }
    return result;
}

// wł閾邳͈̔͂擾
Bool Camera::GetBrightnessRange(SInt16Ptr minimum, SInt16Ptr maximum)
{
    Bool result(false);
    *minimum = *maximum = 0;
    if (_camera != null) {
        Bool support;
        if (_camera->IsBrightness(&support) == SFERR_NO_ERROR) {
            if (support) {
                SInt32 value;
                AEEParmInfo info;
                if (_camera->GetParm(CAM_PARM_BRIGHTNESS, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {    // ͈͂擾
                    *minimum = (SInt16)info.nMin;
                    *maximum = (SInt16)info.nMax;
                    result = true;
                }
            }
        }
    }
    return result;
}

// 邳w肷
Void Camera::SetBrightness(SInt16 value)
{
    if (_camera != null) {
        if (_recordable) {
            _camera->SetBrightness(value);  // 邳w
        }
    }
}

// ݂̃Y[l擾
SInt16 Camera::GetZoom(Void)
{
    SInt16 result(0);
    if (_camera != null) {
        SInt32 value;
        AEEParmInfo info;
        if (_camera->GetParm(CAM_PARM_ZOOM, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {  // Y[l擾
            result = (SInt16)value;
        }
    }
    return result;
}

// włJl͈̔͂擾
Bool Camera::GetZoomRange(SInt16Ptr minimum, SInt16Ptr maximum)
{
    Bool result(false);
    *minimum = *maximum = 0;
    if (_camera != null) {
        Bool support;
        if (_camera->IsZoom(&support) == SFERR_NO_ERROR) {
            if (support) {
                SInt32 value;
                AEEParmInfo info;
                if (_camera->GetParm(CAM_PARM_ZOOM, &value, (SInt32Ptr)&info) == SFERR_NO_ERROR) {  // ͈͂擾
                    *minimum = (SInt16)info.nMin;
                    *maximum = (SInt16)info.nMax;
                    result = true;
                }
            }
        }
    }
    return result;
}

// Y[lݒ肷
Void Camera::SetZoom(SInt16 value)
{
    if (_camera != null) {
        if (_recordable) {
            _camera->SetZoom(value);    // Y[lw
        }
    }
}
