#include "Complex.hpp"

Complex::Complex(Float64 x, Float64 y) : _x(x), _y(y)
{
}

Complex::Complex(ComplexConstRef z) : _x(z._x), _y(z._y)
{
}

Complex::~Complex(Void)
{
}

Float64 Complex::GetReal(Void) const
{
    return _x;
}

Float64 Complex::GetImaginary(Void) const
{
    return _y;
}

Float64 Complex::GetReal(ComplexConstRef z)
{
    return z._x;
}

Float64 Complex::GetImaginary(ComplexConstRef z)
{
    return z._y;
}

ComplexConstRef Complex::SetReal(Float64 x)
{
    _x = x;
    return *this;
}

ComplexConstRef Complex::SetImaginary(Float64 y)
{
    _y = y;
    return *this;
}

Float64ConstRef Complex::Re(Void) const
{
    return _x;
}

Float64Ref Complex::Re(Void)
{
    return _x;
}

Float64 Complex::Re(ComplexConstRef z)
{
    return z._x;
}

Float64ConstRef Complex::Im(Void) const
{
    return _y;
}

Float64Ref Complex::Im(Void)
{
    return _y;
}

Float64 Complex::Im(ComplexConstRef z)
{
    return z._y;
}

Float64 Complex::Abs(Void) const
{
    return sqrt(_x * _x + _y * _y);
}

Float64 Complex::Abs2(Void) const
{
    return _x * _x + _y * _y;
}

Float64 Complex::Abs(ComplexConstRef z)
{
    return sqrt(z._x * z._x + z._y * z._y);
}

Float64 Complex::Abs2(ComplexConstRef z)
{
    return z._x * z._x + z._y * z._y;
}

Complex Complex::GetCC(Void) const
{
    Complex ret(_x, -_y);
    return ret;
}

Complex Complex::GetCC(ComplexConstRef z)
{
    Complex ret(z._x, -z._y);
    return ret;
}

ComplexConstRef Complex::operator=(ComplexConstRef z)
{
    _x = z._x;
    _y = z._y;
    return *this;
}

ComplexConstRef Complex::operator=(Float64 a)
{
    _x = a;
    _y = 0;
    return *this;
}

Complex Complex::operator+(ComplexConstRef z) const
{
    Complex ret(_x + z._x, _y + z._y);
    return ret;
}

Complex Complex::operator+(Float64 a) const
{
    Complex ret(_x + a, _y);
    return ret;
}

Complex operator+(Float64 a, ComplexConstRef z)
{
    Complex ret(a + z._x, z._y);
    return ret;
}

ComplexConstRef Complex::operator+(Void) const
{
    return *this;
}

Complex Complex::operator-(ComplexConstRef z) const
{
    Complex ret(_x - z._x, _y - z._y);
    return ret;
}

Complex Complex::operator-(Float64 a) const
{
    Complex ret(_x - a, _y);
    return ret;
}

Complex operator-(Float64 a, ComplexConstRef z)
{
    Complex ret(a - z._x, -z._y);
    return ret;
}

Complex Complex::operator-(Void) const
{
    Complex ret(-_x, -_y);
    return ret;
}

Complex Complex::operator*(ComplexConstRef z) const
{
    Complex ret(_x * z._x - _y * z._y, _x * z._y + _y * z._x);
    return ret;
}

Complex Complex::operator*(Float64 a) const
{
    Complex ret(_x * a, _y * a);
    return ret;
}

Complex operator*(Float64 a, ComplexConstRef z)
{
    Complex ret(a * z._x, a * z._y);
    return ret;
}

Complex Complex::operator/(ComplexConstRef z) const
{
    Float64 c = fabs(z._x);
    Float64 d = fabs(z._y);
    Float64 g = (c > d) ? c : d;
    ASSERT(g != 0);
    c = z._x / g;
    d = z._y / g;
    g = c * z._x + d * z._y;
    Complex ret((_x * c + _y * d) / g, (_y * c - _x * d) / g);
    return ret;
}

Complex Complex::operator/(Float64 a) const
{
    ASSERT(a != 0);
    Complex ret(_x / a, _y / a);
    return ret;
}

Complex operator/(Float64 a, ComplexConstRef z)
{
    Complex temp(a, 0);
    return temp / z;
}

ComplexConstRef Complex::operator+=(ComplexConstRef z)
{
    _x += z._x;
    _y += z._y;
    return *this;
}

ComplexConstRef Complex::operator+=(Float64 a)
{
    _x += a;
    return *this;
}

ComplexConstRef Complex::operator-=(ComplexConstRef z)
{
    _x -= z._x;
    _y -= z._y;
    return *this;
}

ComplexConstRef Complex::operator-=(Float64 a)
{
    _x -= a;
    return *this;
}

ComplexConstRef Complex::operator*=(ComplexConstRef z)
{
    Complex ret = (*this) * z;
    *this = ret;
    return *this;
}

ComplexConstRef Complex::operator*=(Float64 a)
{
    _x *= a;
    _y *= a;
    return *this;
}

ComplexConstRef Complex::operator/=(ComplexConstRef z)
{
    Complex ret = (*this) / z;
    *this = ret;
    return *this;
}

ComplexConstRef Complex::operator/=(Float64 a)
{
    ASSERT(a != 0);
    _x /= a;
    _y /= a;
    return *this;
}
