#ifndef __MATRIX_H
#define __MATRIX_H

#include <SophiaFramework.hpp>
#include "Array.hpp"

// 2zNX


template <typename T>
class Matrix {
private:
    SInt32 _rowSize;
    SInt32 _columnSize;
    Array<T> _array;

public:
    // RXgN^
    Matrix(Void) : _rowSize(0), _columnSize(0) {
    }

    // Rs[ERXgN^
    Matrix(const Matrix<T> &ob) {
        Set(ob);
    }

    // fBXgN^
    ~Matrix(Void) {
        Clear();
    }

    // TCY擾
    SInt32 GetColumnSize(Void) const {
        return _columnSize;
    }

    // sTCY擾
    SInt32 GetRowSize(Void) const {
        return _rowSize;
    }

    // ׂď
    Void Clear(Void) {
        _array.Clear();
        _columnSize = _rowSize = 0;
    }

    // TCYݒ
    SFCError SetSize(SInt32 row, SInt32 column) {
        ASSERT(row >= 0 && column >= 0);
        SInt32 size(row * column);
        SFCError ret(SFERR_NO_ERROR);

        if (size != _array.GetSize()) {
            ret = _array.SetSize(size);
        }
        if (ret == SFERR_NO_ERROR) {
            _rowSize = row;
            _columnSize = column;
        }
        return ret;
    }

    // sݒ
    SFCError Set(const Matrix<T> &ob) {
        SFCError ret(_array.Set(ob._array));
        if (ret == SFERR_NO_ERROR) {
            _rowSize = ob._rowSize;
            _columnSize = ob._columnSize;
        } else {
            Clear();
        }
        return ret;
    }

    // vf擾
    const T &Get(SInt32 row, SInt32 column) const {
        ASSERT(row >= 0 && column >= 0 && row < _rowSize && column < _columnSize);
        return _array.Get(row * _columnSize + column);
    }

    // vfݒ
    SFCError Set(SInt32 row, SInt32 column, const T &ob) {
        ASSERT(row >= 0 && column >= 0 && row < _rowSize && column < _columnSize);
        return _array.Set(row * _columnSize + column, ob);
    }

    // Zq
    const Matrix<T> &operator =(const Matrix<T> &ob) {
        Set(ob);
        return *this;
    }

    T &operator ()(SInt32 row, SInt32 column) {
        ASSERT(row >= 0 && column >= 0 && row < _rowSize && column < _columnSize);
        return _array.operator [](row * _columnSize + column);
    }

    const T &operator ()(SInt32 row, SInt32 column) const {
        ASSERT(row >= 0 && column >= 0 && row < _rowSize && column < _columnSize);
        return _array.operator [](row * _columnSize + column);
    }
};

#endif
