#include "RecogEngine.hpp"


RecogEngine::RecogEngine(Void) : _processFlag(false), _coordNumber(COORD_NUMBER), _parcorOrder(PARCOR_ORDER)
{
#if !defined(MAKE_PARCOR_TABLE)
    _parcorTableLoaded = LoadParcorTable();
#endif
}

RecogEngine::RecogEngine(RecogEngineConstRef obj) : _processFlag(false)
{
    Copy(obj);
}

RecogEngine::~RecogEngine(Void)
{
    _recogString.Clear();
    _monochrome.Clear();
    _coord.Clear();
    _processFlag = false;
}

Bool RecogEngine::Copy(RecogEngineConstRef obj)
{
    if (_monochrome.Set(obj._monochrome) == SFERR_NO_ERROR && _coord.Set(obj._coord) == SFERR_NO_ERROR && _parcor.Set(obj._parcor) == SFERR_NO_ERROR
        && _parcorTable.Set(obj._parcorTable) == SFERR_NO_ERROR && _charTable.Set(obj._charTable) == SFERR_NO_ERROR) {
        _recogString = obj._recogString;
        _pbitmap = obj._pbitmap;
        _bitmapWidth = obj._bitmapWidth;
        _bitmapHeight = obj._bitmapHeight;
        _coordNumber = COORD_NUMBER;
        _parcorOrder = PARCOR_ORDER;
#if !defined(MAKE_PARCOR_TABLE)
        if (!_parcorTableLoaded) {
            _parcorTableLoaded = LoadParcorTable();
        }
#endif

        return true;
    } else {
        _recogString.Clear();
        _monochrome.Clear();
        _coord.Clear();
        _parcorTable.Clear();
        _charTable.Clear();
        _pbitmap = obj._pbitmap;
        _bitmapWidth = 0;
        _bitmapHeight = 0;
        _coordNumber = COORD_NUMBER;
        _parcorOrder = PARCOR_ORDER;
#if !defined(MAKE_PARCOR_TABLE)
        _parcorTableLoaded = false;
#endif

        return false;
    }
}

Void RecogEngine::MakeGrayScale(Void)
{
    // rbg}bv擾
    _bitmapWidth = _pbitmap->GetWidth();
    _bitmapHeight = _pbitmap->GetHeight();
#if !defined(QUIET)
    TRACE("RecogEngine::MakeGrayScale execute");
#endif
    // z̊m
    _monochrome.Clear();
    if (_monochrome.SetSize(_bitmapHeight, _bitmapWidth) != SFERR_NO_ERROR) {
        _processFlag = false;
        return;
    }

    SInt16 i, j;

#if defined(TARGET_ENVIRONMENT_PHYSICAL)
    SFXGraphicsPtr pg = SFXGraphics::GetInstance();
#endif
    // O[EXP[𐶐
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            // _ (j, i)  RGB l擾
            SFXRGBColor rgb = _pbitmap->GetPixel(j, i);
            // RGB l  255K̔摜ɕϊ
            UInt16 lv = (30 * static_cast<UInt16>(rgb.GetRed()) + 59 * static_cast<UInt16>(rgb.GetGreen()) + 11 * static_cast<UInt16>(rgb.GetBlue())) / 100;
            _monochrome(i, j) = static_cast<UInt08>(lv);
#if defined(TARGET_ENVIRONMENT_PHYSICAL)
            // @ł́Aɕϊꂽ摜\
            pg->SetForeColor(SFXRGBColor(lv, lv, lv, 0x00));
            pg->DrawPixel(SFXPixel(j, i));
#endif
            // ***  ***
            // _monochrome ͍sȂ̂ŁAʏ̍WƂ͈tɂȂ
        }
    }

#if !defined(MAKE_PARCOR_TABLE)
    // mCŶ߂ɃfBAEtB^|
    Matrix<UInt08> median;
    SInt08 k, l, m;

    if (median.SetSize(_bitmapHeight, _bitmapWidth) == SFERR_NO_ERROR) {
        for (i = 0; i < _bitmapHeight; i++) {
            for (j = 0; j < _bitmapWidth; j++) {
                UInt08 val[9];
                for (m = 0; m < 9; m++) {
                    // z̏
                    val[m] = 0;
                }
                for (k = -1; k <= 1; k++) {
                    for (l = -1; l <= 1; l++) {
                        SInt32 u = i + k;
                        SInt32 v = j + l;
                        if (u < 0 || u >= _bitmapHeight || v < 0 || v >= _bitmapWidth) {
                            u = i;
                            v = j;
                        }
                        val[3 * (k + 1) + (l + 1)] = _monochrome(u, v);
                    }
                }
                // 邳̈Â̂珇ɕׂ
                for (k = 0; k < 9; k++) {
                    for (l = k + 1; l < 9; l++) {
                        if (val[k] > val[l]) {
                            UInt08 t = val[k];
                            val[k] = val[l];
                            val[l] = t;
                        }
                    }
                }
                // 5ԖڂɈÂ邳A_ (j, i) ̖邳Ƃ
                median(i, j) = val[4];
            }
        }
        _monochrome = median;
        median.Clear();
    }
#endif

#if !defined(QUIET)
    TRACE("RecogEngine::MakeGrayScale done");
#endif
    _processFlag = true;
}

#define HISTO_MAX 256           // UInt08 ̏ + 1

UInt08 RecogEngine::CalcThreshold(Void)
{
    // O[EXP[摜̐ɎsĂA1Ԃ
    if (!_processFlag) {
        return 1;
    }


    // qXgO̐
    UInt32Ptr pHisto = new UInt32[HISTO_MAX];
    if (pHisto == null) {
        return 1;
    }

    SInt16 i, j, k, l;

    // qXgȌ
    for (i = 0; i < HISTO_MAX; i++) {
        pHisto[i] = 0;
    }

#if !defined(MAKE_PARCOR_TABLE)
    // JŎBe摜gpꍇAmCỶel
    // ֊s̖邳găqXgO\

    // ܂̓vVAEtB^ŗ֊so
    Matrix<UInt08> shape;
    if (shape.SetSize(_bitmapHeight, _bitmapWidth) != SFERR_NO_ERROR) {
        delete [] pHisto;
        return 1;
    }
    SInt08 lap[9] = {1, 1, 1, 1, -8, 1, 1, 1, 1};   // vVA
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            // ۂɃvXZq
            SInt16 val = 0;
            for (k = -1; k <= 1; k++) {
                for (l = -1; l <= 1; l++) {
                    SInt32 u = i + k;
                    SInt32 v = j + l;
                    if (u >= 0 && u < _bitmapHeight && v >= 0 && v < _bitmapWidth) {
                        val += static_cast<SInt16>(lap[3 * (i + 1) + (j + 1)]) * static_cast<SInt16>(_monochrome(u, v));
                    } else {
                        // ͉ʂ̋EȂ̂ŁAȂ
                        val = 0;
                        k = l = 10;
                    }
                }
            }
            if (val < 0) {  // Βl
                val = -val;
            }
            shape(i, j) = static_cast<UInt16>(val) >> 3;        // >> 3  / 8 Ɠ
        }
    }

    // oꂽ֊s̒猰Ȃ̂߂
    // x̏ 30% ̗p
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            pHisto[shape(i, j)]++;
        }
    }
    UInt32 s = (3 * _bitmapHeight / 10) * _bitmapWidth;
    UInt32 sum = 0;
    UInt08 shapeThreshold = 0;
    for (i = HISTO_MAX - 1; i >= 0; i--) {
        sum += pHisto[i];
        if (sum > s) {
            if (i == HISTO_MAX - 1) {
                shapeThreshold = static_cast<UInt08>(i);
            } else {
                shapeThreshold = static_cast<UInt08>(i + 1);
            }
            i = -10;
        }
    }

    // ̏̂߂ɁAqXgO͏Ă
    for (i = 0; i < HISTO_MAX; i++) {
        pHisto[i] = 0;
    }
    sum = 0;
#endif

    Float64 f = 0;  // 摜Ŝ̋Pxςi[
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
#if !defined(MAKE_PARCOR_TABLE)
            if (shape(i, j) >= shapeThreshold) {
                sum++;
#endif
                pHisto[_monochrome(i, j)]++;
                f += static_cast<Float64>(_monochrome(i, j));
#if !defined(MAKE_PARCOR_TABLE)
            }
#endif
        }
    }
#if defined(MAKE_PARCOR_TABLE)
    f /= static_cast<Float64>(_bitmapWidth * _bitmapHeight);
#else
    f /= static_cast<Float64>(sum);
#endif

#if !defined(QUIET)
    // qXgOfobOEEBhEɏo
    for (i = 0; i < HISTO_MAX; i++) {
#if defined(DEVICE_TEST)
        DBGPRINTF("HISTO[%d] = %d", i, pHisto[i]);
#else
        TRACE("HISTO[%d] = %d", i, pHisto[i]);
#endif
    }
    TRACE("SIZE = %d x %d", _bitmapWidth, _bitmapHeight);
#endif

    // ʕ͖@ɂ臒l̎Zo
    Float64 covRMax = 0;
    UInt08 threshold = 0;
    SInt16 t;
    for (t = 0; t < HISTO_MAX; t++) {
        SInt16 i;
        UInt32 n1 = 0, n2 = 0;
        Float64 f1 = 0, f2 = 0;
        for (i = 0; i < t; i++) {
            n1 += pHisto[i];
            f1 += static_cast<Float64>(i) * static_cast<Float64>(pHisto[i]);
        }
        Float64 n1F = static_cast<Float64>(n1);
        f1 /= n1F;
        for (i = t; i < HISTO_MAX; i++) {
            n2 += pHisto[i];
            f2 += static_cast<Float64>(i) * static_cast<Float64>(pHisto[i]);
        }
        Float64 n2F = static_cast<Float64>(n2);
        f2 /= n2F;

        Float64 d1 = f1 - f;
        Float64 d2 = f2 - f;

        // covB : NXԕU2AcovW : NXU2 (ɂꂼ̕q̂)
        Float64 covB = n1F * d1 * d1 + n2F * d2 * d2;
        Float64 covW = 0;
        for (i = 0; i < t; i++) {
            d1 = static_cast<Float64>(i) - f1;
            covW += static_cast<Float64>(pHisto[i]) * d1 * d1;
        }
        for (i = t; i < HISTO_MAX; i++) {
            d2 = static_cast<Float64>(i) - f2;
            covW += static_cast<Float64>(pHisto[i]) * d2 * d2;
        }

        Float64 covR = covB / covW;
        if (covR >= covRMax) {
            covRMax = covR;
            threshold = static_cast<UInt08>(t);
        }
    }

    delete [] pHisto;
    return threshold;
}

#define TRUE_BITS   0x80        // ̈ꕔł
#define FALSE_BITS  0x00        // wïꕔł

Void RecogEngine::MakeMonochrome(Void)
{
    // 2ls
    // 臒lȉł()̒l TRUE_BITS ɒuA
    // 臒lȏł()̒l FALSE_BITS ɒu

    if (!_processFlag) { // ϊɎsĂ
        return;
    }

#if !defined(QUIET)
    TRACE("RecogEngine::MakeMonochrome execute");
#endif

    UInt08 threshold = CalcThreshold();
#if !defined(QUIET)
#if defined(DEVICE_TEST)
    DBGPRINTF("Threshold = %d", threshold);
#else
    TRACE("Threshold = %d", threshold);
#endif
#endif

    for (SInt16 i = 0; i < _bitmapHeight; i++) {
        for (SInt16 j = 0; j < _bitmapWidth; j++) {
            if (_monochrome(i, j) < threshold) {
                _monochrome(i, j) = TRUE_BITS;
            } else {
                _monochrome(i, j) = FALSE_BITS;
            }
        }
    }

#if !defined(QUIET)
    TRACE("RecogEngine::MakeMonochrome done");
#endif
    _processFlag = true;
}

Void RecogEngine::GetOutline(Void)
{
    if (!_processFlag) {
        return;
    }
#if !defined(QUIET)
    TRACE("RecogEngine::GetOutline begin");
#endif
    SInt16 i, j, k, l;

    // 8}X̒ɓ_邩
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            for (k = -1; k <= 1; k++) {
                for (l = -1; l <= 1; l++) {
                    SInt32 u = i + k;
                    SInt32 v = j + l;
                    if (u >= 0 && u < _bitmapHeight && v >= 0 && v < _bitmapWidth
                        && (k != 0 || l != 0)) {
                        UInt08 f = _monochrome(u, v);
                        if (f & TRUE_BITS) {
                            f = _monochrome(i, j) + 1;
                            _monochrome(i, j) = f;
                        }
                    }
                }
            }
        }
    }

    // 4 - 6 }X̗̂̂p
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            UInt08 n = _monochrome(i, j) & (~TRUE_BITS);
            if (n >= 4 && n <= 6) {
                _monochrome(i, j) = 1;
            } else {
                _monochrome(i, j) = 0;
            }
        }
    }
#if !defined(QUIET)
    TRACE("RecogEngine::GetOutline done");
#endif
    _processFlag = true;
}

Void RecogEngine::GetCoordinatePoints(Void)
{
    _coord.Clear();
    if (!_processFlag) {
        return;
    }
#if !defined(QUIET)
    TRACE("RecogEngine::GetCoordinatePoints execute");
#endif

    // X^[gn_T
    Array<SFXPixel> coordWork;
    SInt16 i, j, iback = 0, jback = 0;
#if ((defined(EMULATOR_TEST) && defined(DRAW_CHAR)) || !defined(MAKE_PARCOR_TABLE))
    SFXGraphicsPtr pg = SFXGraphics::GetInstance();
    pg->SetForeColor(SFXRGBColor(0x00, 0x00, 0xFF, 0x00));
    for (i = 0; i < _bitmapHeight; i++) {
        for (j = 0; j < _bitmapWidth; j++) {
            if (_monochrome(i, j)) {
                pg->DrawPixel(SFXPixel(j, i));
            }
        }
    }
#endif
    for (i = _bitmapHeight - 1; i >= 0; i--) {
        for (j = 0; j < _bitmapWidth; j++) {
            if (_monochrome(i, j)) {
                if (coordWork.Append(SFXPixel(j, i)) != SFERR_NO_ERROR) {
                    _processFlag = false;
                    return;
                }
                _monochrome(i, j) = 0;
                // [v甲
                iback = i;
                jback = j;
                i = -1;
                j = _bitmapWidth;
            }
        }
    }

    i = iback;
    j = jback;
#if ((defined(EMULATOR_TEST) && defined(DRAW_CHAR)) || !defined(MAKE_PARCOR_TABLE))
    pg->SetForeColor(SFXRGBColor(0xFF, 0x00, 0x00, 0x00));
#endif
    // ֊s1
    while (true) {
#if ((defined(EMULATOR_TEST) && defined(DRAW_CHAR)) || !defined(MAKE_PARCOR_TABLE))
        pg->DrawPixel(SFXPixel(j, i));
#endif
        SInt16 k[8] = { 0,  0, -1,  1, -1, -1,  1,  1};
        SInt16 l[8] = {-1,  1,  0,  0,  1, -1, -1,  1};
        SInt08 m;
        for (m = 0; m < 8; m++) {
            SInt32 u = i + k[m];
            SInt32 v = j + l[m];
            if (u >= 0 && u < _bitmapHeight && v >= 0 && v < _bitmapWidth) {
                if (_monochrome(u, v)) {        // ׂ̗֊s_
                    i = static_cast<SInt16>(u);
                    j = static_cast<SInt16>(v);
                    if (coordWork.Append(SFXPixel(j, i)) != SFERR_NO_ERROR) {
                        _processFlag = false;
                        return;
                    }
                    _monochrome(u, v) = 0;
                    m = 20; // ׂ̗֊s_JԂ
                }
            }
        }
        if (m < 20) {       // ֊s_̑
            break;
        }
    }

    SInt32 count = coordWork.GetSize();
#if !defined(QUIET)
    TRACE("total coord = %d", count);
#endif
    // ő _coordNumber ̒_̗p
    if (count > _coordNumber) {
        Float64 s = static_cast<Float64>(count) / static_cast<Float64>(_coordNumber);
        Array<SFXPixel> work;
        if (work.SetSize(_coordNumber) != SFERR_NO_ERROR) {
            _processFlag = false;
            return;
        }
        for (i = 0; i < _coordNumber; i++) {
            Float64 i_f = static_cast<Float64>(i);
            SInt32 k = static_cast<SInt32>(floor(i_f * s + 0.5));
            work[i] = coordWork[k];
        }
        coordWork = work;
        work.Clear();
        count = _coordNumber;
    }

    // ֊s_̏dSƂ߂āAꂼ̍WĂ
#if !defined(QUIET)
    TRACE("coord = %d", count);
#endif
#if ((defined(EMULATOR_TEST) && defined(DRAW_CHAR)) || !defined(MAKE_PARCOR_TABLE))
    pg->SetForeColor(SFXRGBColor(0x00, 0xFF, 0x00, 0x00));
#endif
    UInt32 center_x = 0, center_y = 0;
    for (i = 0; i < count; i++) {
        SFXPixel point = coordWork[i];
        center_x += point.GetX();
        center_y += point.GetY();
#if !defined(QUIET)
        TRACE("coord (%d, %d)", point.GetX(), point.GetY());
#endif
#if ((defined(EMULATOR_TEST) && defined(DRAW_CHAR)) || !defined(MAKE_PARCOR_TABLE))
        pg->DrawPixel(point);
#endif
    }
    Float64Const n(static_cast<Float64>(count));
    Complex center(static_cast<Float64>(center_x) / n, static_cast<Float64>(center_y) / n);
    if (_coord.SetSize(count) != SFERR_NO_ERROR) {
        _processFlag = false;
        return;
    }
    for (i = 0; i < count; i++) {
        SFXPixel point = coordWork[i];
        _coord[i] = Complex(point.GetX(), point.GetY()) - center;
    }
    _monochrome.Clear();    // 摜͈ȌKvȂ
#if !defined(QUIET)
    TRACE("RecogEngine::GetCoordinatePoints done");
#endif
    _processFlag = true;
}

Void RecogEngine::CalcParcor(Void)
{
    _parcor.Clear();
    if (!_processFlag) {
        return;
    }
#if !defined(QUIET)
    TRACE("RecogEngine::CalcParcor execute");
#endif

    SInt32 coordNum = _coord.GetSize();
#if !defined(QUIET)
    TRACE("coordNum = %d", coordNum);
#endif
    SInt32 coordNumInc = coordNum + 1;

    // f֌W r[i] ߂
    Array<Complex> r;           // f֌W
    if (r.SetSize(coordNumInc) != SFERR_NO_ERROR) {
        _processFlag = false;
        return;
    }

    SInt16 i, j;

    for (i = 0; i <= _parcorOrder; i++) {
        // r[i] ̌vZ
        Complex sum(0, 0);
        for (j = 0; j < coordNum; j++) {
            sum += _coord[j] * Complex::GetCC(_coord[(i + j) % coordNum]);
        }
        r[i] = sum / coordNum;
    }
#if !defined(QUIET)
    TRACE("PARCOR calculation done");
#endif
    _coord.Clear();     // Ȍsv


    // f PARCOR WvZ(CfbNX1)
    Array<Complex> a1, a2;
    if (_parcor.SetSize(coordNumInc) != SFERR_NO_ERROR || a1.SetSize(coordNumInc) != SFERR_NO_ERROR || a2.SetSize(coordNumInc) != SFERR_NO_ERROR) {
        TRACE("coordNumInc = %d", coordNumInc);
        _parcor.Clear();
        _processFlag = false;
        return;
    }

    _parcor[0] = a1[0] = 0;         // ɈӖ͂Ȃ
    _parcor[1] = a1[1] = r[1] / r[0];
    for (i = 2; i <= _parcorOrder; i++) {
        Complex parcorNum = r[i];
        Complex parcorDen = r[0];
        for (j = 1; j < i; j++) {
            parcorNum -= r[i - j] * a1[j];
            parcorDen -= r[j] * Complex::GetCC(a1[j]);
        }
        a2[i] = _parcor[i] = parcorNum / parcorDen;
        for (j = 1; j < i; j++) {
            a2[j] = a1[j] - Complex::GetCC(a1[i - j]) * _parcor[i];
        }
        a1 = a2;
    }

    r.Clear();
    a1.Clear();
    a2.Clear();
#if !defined(QUIET)
    TRACE("RecogEngine::CalcParcor done");
#endif
    _processFlag = true;
}

#if defined(MAKE_PARCOR_TABLE)
Bool RecogEngine::MakeParcorTable(Void)
{
    // \[X猳ƂȂ镶̉摜ǂݍłꂼ̕f PARCOR WvZAt@Cɕۑ
    _parcorTable.Clear();
    _charTable.Clear();
    if (_parcorTable.SetSize(TABLE_SIZE, _parcorOrder) != SFERR_NO_ERROR || _charTable.SetSize(TABLE_SIZE) != SFERR_NO_ERROR) {
        _parcorTable.Clear();
        _charTable.Clear();
        return false;
    }
    SInt16 count = 0;

#if !defined(QUIET)
    TRACE("RecogEngine::MakeParcorTable execute");
#endif
    for (SInt16 i = 0; i < 26; i++) {
        // At@xbgE啶e[u
        Bitmap bitmap;
        bitmap.LoadResBitmap(CHARS_RES_FILE, BITMAP_LARGE_A + i);
        _pbitmap = &bitmap;
#if !defined(QUIET)
        TRACE("\'%c\' begin", 'A' + i);
#endif
        MakeGrayScale();
        MakeMonochrome();
        GetOutline();
        GetCoordinatePoints();
        CalcParcor();
        if (_processFlag) {
            _charTable[count] = 'A' + i;
            for (SInt16 j = 0; j < _parcorOrder; j++) {
                _parcorTable(count, j) = _parcor[j + 1];
            }
            count++;
            TRACE("\'%c\' success", 'A' + i);
        }
    }

    return true;
}
#endif

#if defined(MAKE_PARCOR_TABLE)
Bool RecogEngine::SaveParcorTable(Void)
{
#if !defined(QUIET)
    TRACE("PARCOR table save start");
#endif
    SFXFile::Remove(SFXPath(PARCOR_TABLE_NAME));

    // t@C̍쐬
    SFXFile file;
    SFXBinaryStreamWriter writer;
    SFCError err;

    err = file.OpenReadWrite(SFXPath(PARCOR_TABLE_NAME));
    if (err != SFERR_NO_ERROR) {
#if !defined(QUIET)
        TRACE("failed create %s (error code = %d)", PARCOR_TABLE_NAME, err);
#endif
        return false;
    }
    err = file.GetStreamWriter(&writer);
    if (err != SFERR_NO_ERROR) {
        return false;
    }

    // PARCOR We[ũtH[}bg
    // 1. o^Ă镶̎      SInt16
    // 2. ֊s_̐                 SInt16
    // 3. PARCOR W̎              SInt16
    // 4.                             WChar
    // 5.                             Float64
    // 6.                             Float64
    // ȍ~ 4 ~ 6 ̌JԂ

    SInt16 n = static_cast<SInt16>(_charTable.GetSize());

    writer << n << _coordNumber << _parcorOrder;
    for (SInt16 i = 0; i < n; i++) {
        WChar ch = _charTable[i];

        writer << ch;
        for (SInt16 j = 0; j < _parcorOrder; j++) {
            Complex z = _parcorTable(i, j);

            writer << z.Re() << z.Im();
        }
    }
    err = writer.Flush();
    if (err != SFERR_NO_ERROR) {
        return false;
    }
    writer.Release();
    file.Close();

#if !defined(QUIET)
    TRACE("PARCOR table save done");
#endif
    return true;
}
#endif

Bool RecogEngine::LoadParcorTable(Void)
{
#if !defined(QUIET)
    TRACE("PARCOR table load start");
#endif
    SFXFile file;
    SFXBinaryStreamReader reader;
    SFCError err;

    err = file.OpenReadOnly(SFXPath(PARCOR_TABLE_NAME));
    if (err != SFERR_NO_ERROR) {
#if !defined(QUIET)
        TRACE("%s load failed (error code = %d)", PARCOR_TABLE_NAME, err);
#endif
        DBGPRINTF("Error %s", PARCOR_TABLE_NAME);
        return false;
    }
    err = file.GetStreamReader(&reader);
    if (err != SFERR_NO_ERROR) {
        return false;
    }
    err = reader.Fetch();
    if (err != SFERR_NO_ERROR) {
        return false;
    }

    SInt16 n;

    reader >> n >> _coordNumber >> _parcorOrder;
    _parcorTable.Clear();
    _charTable.Clear();
    if (_parcorTable.SetSize(n, _parcorOrder) != SFERR_NO_ERROR || _charTable.SetSize(n) != SFERR_NO_ERROR) {
        return false;
    }
    for (SInt16 i = 0; i < n; i++) {
        WChar ch;

        reader >> ch;
        _charTable[i] = ch;
        for (SInt16 j = 0; j < _parcorOrder; j++) {
            Float64 t[2];

            reader >> t[0] >> t[1];
            _parcorTable(i, j) = Complex(t[0], t[1]);
        }
    }
    reader.Release();
    file.Close();

#if !defined(QUIET)
    TRACE("PARCOR table load done");
#endif
    return true;
}

Bool RecogEngine::Recognize(SFBBitmapSmpConstRef pbitmap)
{
    if (pbitmap == null) {
        return false;
    }

    _recogString.Clear();
#if !defined(MAKE_PARCOR_TABLE)
    if (!_parcorTableLoaded) {
        return false;
    }
#endif
    _pbitmap = new Bitmap(pbitmap);
    MakeGrayScale();
    MakeMonochrome();
    GetOutline();
    GetCoordinatePoints();
    CalcParcor();
    if (!_processFlag) {
        // PARCOR W̌vZɎs
        delete _pbitmap;
        return false;
    }


    // ŏ2@ɂAłT
    Array<Float64> errs;
    Array<SInt16> list;
    SInt16 n = (SInt16)_charTable.GetSize();
    if (errs.SetSize(n) != SFERR_NO_ERROR || list.SetSize(n) != SFERR_NO_ERROR) {
        // ŝߔrłȂ
        delete _pbitmap;
        return false;
    }

    Float64 e = 0;
    SInt16 i, j;
    for (i = 0; i < _parcorOrder; i++) {
        Complex t = _parcorTable(0, i) - _parcor[i + 1];
        e += t.Abs2();
    }
    errs[0] = e;
    list[0] = 0;
#if !defined(QUIET)
    WChar buf[30];
    SFXAnsiString str;
    SFXHelper::floattowstr(e, buf, sizeof(buf));
    str = buf;
    TRACE("%c : error = %s", _charTable[0], str.GetCString());
#endif
    for (i = 1; i < n; i++) {
        e = 0;
        for (j = 0; j < _parcorOrder; j++) {
            Complex t = _parcorTable(i, j) - _parcor[j + 1];
            e += t.Abs2();
        }
#if !defined(QUIET)
        SFXHelper::floattowstr(e, buf, sizeof(buf));
        str = buf;
        TRACE("%c : error = %s", _charTable[i], str.GetCString());
#endif
        errs[i] = e;
        list[i] = i;
    }

    // 덷̂珇ɕׂ
    for (i = 0; i < n; i++) {
        for (j = i + 1; j < n; j++) {
            SInt16 a = list[i];
            SInt16 b = list[j];
            if (errs[a] > errs[b]) {
                list[i] = b;
                list[j] = a;
            }
        }
    }

    _recogString = SFXWideString::Format("1.%c 2.%c 3.%c", _charTable[list[0]], _charTable[list[1]], _charTable[list[2]]);

    delete _pbitmap;
    return true;
}

SFXWideStringConstRef RecogEngine::GetString(Void) const
{
    return _recogString;
}
